### include guard
(return 0 2>/dev/null)        &&
[[ -z ${FULLUP_ARGS_SH_+1} ]] &&
declare -xr FULLUP_ARGS_SH_=1 && {
###

# Properly, completely reset and re-init the variables we need. (Yes, there
# are two unset commands for two specific reasons). Command chaining causes
# the program to abort if any variable has the readonly attribute.
__args_init_globals() {
	unset   -n     PROG NREPTS \
	             OPT_FLAG OPT_ID \
	           OPT_ARGFLAGS OPT_REPEATFLAGS FLAGS \
	         LONG_OPT OPT_CHAR OPT_DESC OPT_HELP GLOB ARGS &&
	unset  LONG_OPT OPT_CHAR OPT_DESC OPT_HELP GLOB ARGS \
	        OPT_ARGFLAGS OPT_REPEATFLAGS FLAGS \
	          OPT_FLAG OPT_ID \
	            PROG NREPTS &&
	declare -g   PROG &&
	declare -gai  NREPTS &&
	declare -gAi   OPT_FLAG OPT_ID &&
	declare -g -i    OPT_ARGFLAGS OPT_REPEATFLAGS FLAGS &&
	declare -g -a      LONG_OPT OPT_CHAR OPT_DESC OPT_HELP GLOB ARGS
}

# program_args() - Configure the program's name, command-line arguments and
# command help texts.
#
# This function sets up the program's global name, command-line options and
# help. The program name is written to a variable named PROG and the option
# definitions to two arrays named OPT_CHAR and LONG_OPT. The help texts are
# saved to the OPT_HELP array, and a descriptions of the options' arguments
# are written to OPT_DESC.
#
# The first argument to program_args() must be the program name. It must be
# followed by one or more triplets of arguments.
#
# Each triplet's first field defines the command-line option, the 2nd field
# describes its argument (if any), and the last field contains a short help
# text which must be properly quoted to prevent word splitting.
#
# If a triplet's first field doesn't end in ":" or ":+" the second field is
# ignored and should be set to the empty string. All fields must be present
# or parsing will fail.
#
# An example:
#
#   program_args example-tool 'f--file:' '<path>' 'the input file' \
#                             'q--quiet' ''       ''               \
#                             '--stop'   ''       'stop something'
#
# Each triplet's first field (the option definition) consist of, in order:
#
#   - a short option character (except ':', '+', or '-'), and/or
#   - a long option prefixed by "--" and not followed by ':', '+', or '-',
#   - ':' iff the option requires an argument, and
#   - '+' iff the option can be given multiple times.
#
# The following extended regular expression matches the set of all valid
# option definitions:
#
#   ^([^:\+-]?--[^:\+-][^:\+]*|[^:\+-]):?\+?$
#
# This regexp may in fact be a superset of the valid set, so it's possible
# that it matches weird corner cases which would confuse the parser. It's
# highly recommended to stick to plain ASCII alphanumerics.
#
#   | option        | opt   | long          | needs | allow |
#   | definition    | char  | option        | arg?  | many? |
#   +---------------+-------+---------------+-------+-------+
#   | 9             | -9    |               |       |       |
#   | v+            | -v    |               |       | yes   |
#   | --version     |       | --version     |       |       |
#   | --not+        |       | --not         |       | yes   |
#   | h--help       | -h    | --help        |       |       |
#   | o--or+        | -o    | --or          |       | yes   |
#   | O:            | -O    |               | yes   |       |
#   | U:+           | -U    |               | yes   | yes   |
#   | --run-test:   |       | --run-test    | yes   |       |
#   | --skip:+      |       | --skip        | yes   | yes   |
#   | t--target:    | -t    | --target      | yes   |       |
#   | i--import:+   | -i    | --import      | yes   | yes   |
#   +---------------+-------+---------------+-------+-------+
#
program_args() {
	__args_init_globals && [[ "$1" ]] &&
	(( ($# > 3) && (($# % 3) == 1) )) &&
	{
		readonly PROG="$1"
		shift

		local f0 f1 f2 x
		local -a opt_def

		# deinterleave arguments into 3 arrays
		printf -v f0 '%q%.s%.s\n' "$@"
		printf -v f1 '%.s%q%.s\n' "$@"
		printf -v f2 '%.s%.s%q\n' "$@"

		(( ! ${IFS+1} )) || printf -v x '%q' "$IFS"
		eval "IFS=\$'\n' \
		; opt_def=($f0)  \
		; OPT_DESC=($f1) \
		; OPT_HELP=($f2) \
		${x+; IFS=${x}}"

		x=("${opt_def[@]##--*}")
		OPT_CHAR=("${x[@]/--*[^:\+]}")

		x=("${opt_def[@]#[^-]}")
		x=("${x[@]#[:\+]*}")
		LONG_OPT=("${x[@]#--}")
	}
}

parse_args() {
	local x a parse_err

	__gen_getopt_args x                        &&
	unset -n GETOPT_COMPATIBLE POSIXLY_CORRECT &&
	unset    GETOPT_COMPATIBLE POSIXLY_CORRECT &&
	{
		a="$(getopt "${x[@]}" -- "$@")" || {
			(( $? == 1 )) && parse_err=1
		}
	} && __gen_parser_func __really_parse_args || {
		# internal error
		exit 1
	}

	(( ! parse_err )) && {
		eval "__really_parse_args $a"
	}
}

# Check if the program arguments contained option $1.
#
have_arg() {
	(( 0 != (FLAGS & OPT_FLAG["$1"]) ))
}

# Check if every argument passed to the program was option $1.
#
# The result is meaningful only if $1 has been defined in a program_args()
# call.
#
# The number of times $1 occurs does not affect the outcome; the result is
# true simply if $1 appears at least once and other options are absent.
#
is_lone_arg() {
	(( FLAGS == OPT_FLAG["$1"] ))
}

# Check if no arguments were passed to the program.
#
have_no_args() {
	(( ! FLAGS ))
}

has_optarg() {
	(( 0 != (OPT_ARGFLAGS & OPT_FLAG["$1"]) ))
}

can_repeat() {
	(( 0 != (OPT_REPEATFLAGS & OPT_FLAG["$1"]) ))
}

# usage: get_optarg --option dest_var
get_optarg() {
	set - "${OPT_ID[$1]}" "$2"    &&
	[[ "$1" && "${ARGS[$1]+x}" ]] &&
	eval "$2=(${ARGS[$1]# })"
}

# Get the number of times an option occurred in the program arguments (can
# be 0).
#
# @param $1: option string
# @param $2: result variable
#
# Returns true iff $1 is a defined option and $2 is a valid variable name,
# otherwise false.
#
# A true return value indicates that the result variable has been assigned a
# non-negative integer value. More specifically, on success the result cannot
# be the null string.
#
# A false return implies the result variable has not been accessed in any way.
#
get_arg_count() {
	[[ "$1" && "$2" ]] && {
		set - "${OPT_ID[${1}]}" "$2"
		[[ "$1" ]]
	} && {
		set - "${NREPTS[$1]:-0}" "$2"
		eval "$2=${1@Q}"
	}
}

__delete_null_args() {
	set - "$1" "${@@Q}"
	set - "${@#\'\'}"
	eval "$1=(" "${@:3}" ')'
}

get_long_opts() {
	set - "$1" "${LONG_OPT[@]%+}"
	set - "${@%:}"
	__delete_null_args "$1" "${@:2}"
}

# Generate an array containing option help.
#
# Two array entries per option:
#
#   [0]="-x, --xxx=yyy"  [1]="the x option does x"
#   [2] ...
#
get_help() {
	local -i n="${#GLOB[@]}" i m p q
	local j k out

	for ((i = 0, m = 1, p = 0; i < n; i++, m <<= 1)); do
		k="${GLOB[i]/\|/, }"
		k="${k/#--/    --}"
		[[ -z "$k" ]] || {
			(( ! (OPT_ARGFLAGS & m) )) || {
				j="${OPT_DESC[i]}"
				k+="${j:+=$j}"
			}
			((q = ${#k}, p = (q > p) ? q : p))
			j="${OPT_HELP[i]}"
			if (( (OPT_REPEATFLAGS & m) && ${#j} )); then
				j+=' (can appear more than once)'
			fi
			out+="${k@Q} ${j@Q} "
		}
	done

	{ [[ -z "$2" ]] || eval "$2=$p"; } &&
	eval "$1=($out)"
}

# Generate command-line args for getopt and save them in an array. The first
# (and only) argument is the destination array variable name.
__gen_getopt_args() {
	[[ "$PROG" ]] && (( ${#OPT_CHAR[@]} > 0 )) &&
	(( ${#OPT_CHAR[@]} == ${#LONG_OPT[@]} ))   &&
	local "__$1"                               &&
	{
		# Init the destination array, and declare a temp variable with
		# a name that's guaranteed to not shadow the destination.
		set - "__$1"

		# Append the long options (if any).
		printf -v "$1" -- ',%s' ${LONG_OPT[@]%\+}
		set - "$1" "${!1:1}"

		# Append the short options (if any).
		printf -v "$1" -- '%s' "${OPT_CHAR[@]%\+}"
		set - "${1:2}" "$2" "${!1}"

		eval "$1=("${PROG:+"-n${PROG@Q} "}${2:+"-l${2@Q} "}"-o-${3@Q})"
	}
}

# @param $1 what to name the generated function
__gen_parser_func() {
	local switches key opt dup glob long_opt x
	local -i n i m

	# Generate case statement lines for the options

	OPT_ARGFLAGS=0
	n="${#LONG_OPT[@]}"
	switches=''

	for ((i = 0, m = 1; i < n; i++, m <<= 1)); do
		key="${OPT_CHAR[i]}"

		if [[ "$key" ]]; then
			opt="-$key"
			key="${opt%\+}"
			key="${key%:}"
			[[ -z "${OPT_ID[$key]}${OPT_FLAG[$key]}" ]] || {
				dup="$key"
				break
			}
			OPT_ID["$key"]="$i"
			OPT_FLAG["$key"]="$m"
		fi

		glob="$key"
		long_opt="${LONG_OPT[i]}"

		if [[ "$long_opt" ]]; then
			opt="--$long_opt"
			key="${opt%\+}"
			key="${key%:}"
			[[ -z "${OPT_ID[$key]}${OPT_FLAG[$key]}" ]] || {
				dup="$key"
				break
			}
			OPT_ID["$key"]="$i"
			OPT_FLAG["$key"]="$m"
			glob+="${glob:+|}$key"
		fi

		# this shouldn't happen if all option definitions are valid
		[[ "$key" ]] || continue

		GLOB[i]="$glob"
		opt="${opt#$key}"

		if [[ "${opt: -1}" == '+' ]]; then
			((OPT_REPEATFLAGS |= m))
			printf -v x '\t\t%s) ((p = 0, NREPTS[%d]++, seen |= 0x%x))' "$glob" "$i" "$m"
		else
			printf -v x '\t\t%s) ((NREPTS[%d]++, seen |= 0x%x))' "$glob" "$i" "$m"
		fi

		[[ "${opt:0:1}" != ':' ]] || {
			x+="; ARGS[$i]+=\" \${1@Q}\"; shift"
			((OPT_ARGFLAGS |= m))
		}

		switches+="$x"$' ;;\n'
	done

	[[ -z "$dup" ]] || {
		printf "option '%s' defined multiple times\n" "${dup@Q}" >&2
		return 1
	}

	readonly GLOB OPT_ARGFLAGS OPT_REPEATFLAGS || return 1

	# Generate the function source
	printf -v x '%s\n' \
	"$1"' () {' \
	'	local x' \
	'	local -i seen=0 p=0' \
	'	while (( $# )); do' \
	'		x="$1"; shift' \
	'		case "$x" in' \
	"$switches"'		--) break ;;' \
	'		*)  printf '"'$PROG: unrecognized option %s\n'"' "${x@Q}" >&2; return 1 ;;' \
	'		esac' \
	'		(( seen != p )) || {' \
	'			printf '"'$PROG: too many instances of option %s\n'"' "${x@Q}" >&2' \
	'			return 1' \
	'		}' \
	'		p="$seen"' \
	'	done' \
	'	FLAGS="$seen"' \
	'}'

	# Define the function
	eval "$x"
}

### include guard
} || ((1))
###
