### include guard
(return 0 2>/dev/null)       &&
[[ -z ${FULLUP_CFG_SH_+1} ]] &&
declare -xr FULLUP_CFG_SH_=1 && {
###

# We use "__fullup_cfg_" as our config variable prefix, and we often generate
# variable names dynamically by concatenating it with a suffix. If the suffix
# is accidentally left empty then variable name collisions are inevitable.
# "Reserving" the prefix by declaring it as a readonly variable effectively
# behaves like an assertion guarding against that.
readonly __fullup_cfg_

. "$I/args.sh"

# Unset one or more config variables.
unset_cfg() {
	set - "${@/#/__fullup_cfg_}"
	unset -n "$@" &&
	unset    "$@"
}

# Unset and (re)declare one or more config variables.
declare_cfg() {
	unset_cfg "${@}" &&
	local -ag "${@/#/__fullup_cfg_}"
}

# Unset and (re)initialize a config variable.
init_cfg() {
	unset -n  "__fullup_cfg_$1" &&
	unset     "__fullup_cfg_$1" &&
	local -ag "__fullup_cfg_$1"'=("${@:2}")'
}

# Create (or completely reset) the config variable $1, and initialize it with
# the optargs of command-line argument $2.
#
# Usage:  init_cfg_from_optarg destvar --option
#
init_cfg_from_optarg() {
	set - __fullup_cfg_"$@" &&
	unset -n           "$1" &&
	unset              "$1" &&
	local -ag          "$1" &&
	get_optarg    "$2" "$1"
}

# Like init_cfg_from_optarg(), but create two config variables instead of one
# by treating every optarg as a string of (at most 2) colon-separated values.
#
# Usage:
#
#   init_cfg_from_optarg_split leftvar rightvar --option
#
# For example, if a script that accepts multiple "--xx=..." options is passed
# the arguments
#
#   --xx="A:B" --xx=":" --xx="no_colon" --xx=":no_left_side"
#
# and it then calls
#
#   init_cfg_from_optarg_split lhs rhs --zz
#
# the result is:
#
#   __fullup_cfg_lhs=([0]="A" [1]="" [2]="no_colon" [3]="")
#   __fullup_cfg_rhs=([0]="B" [1]="" [2]=""         [3]="no_left_side")
#
init_cfg_from_optarg_split() {
	local -a a                    &&
	local -i i                    &&
	local    x                    &&
	get_optarg "$3" a             &&
	set - "$2" "$1" "${a[@]%%:*}" &&
	for x in "${@:3}"; do
		x="${a[i]#$x}"
		a[i++]="$x"
	done                          &&
	init_cfg "$1" "${a[@]#:}"     &&
	init_cfg "${@:2}"
}

# This function is used inside platform configuration functions for adding
# configuration items. Previously added items are never overwritten, only
# empty strings are filled in with the new items passed to this function.
#
# An example:
#
#   1. set_cfg  x  1 '' 3           ->  __fullup_cfg_x=("1" "" "3")
#   2. set_cfg  x  a b c            ->  __fullup_cfg_x=("1" "b" "3")
#   3. set_cfg  x  add a few words  ->  __fullup_cfg_x=("1" "b" "3" "words")
#
#
set_cfg() {
	# Make a local copy of the destination array.
	local var="__fullup_cfg_$1[@]"
	local -a arr=("${!var}")
	local -i len="${#arr[@]}" i

	var="__fullup_cfg_$1"
	shift

	# Replace each null string in the destination array with the value
	# located at that same index of the input array.
	for ((i = 0; i < len; i++)); do
		[[ "${arr[i]}" ]] || arr[i]="$1"
		shift
	done

	# If the input array is longer than the destination append the
	# tail end of the input array to the destination.
	(( $# < 1 )) || arr+=("$@")

	set - "${arr[@]@A}"
	var="$var=${3#*=}"
	declare -ag "$var"
}

# This must only be called at the end of a platform configuration function,
# i.e. functions named __fullup_configure_<arch-dep-level>_<target>(). The
# argument passed to __configure() must be the next arch-dep-level upwards.
# Look under plat/ and arch/ to see examples.
#
__configure() {
	[[ "$1" ]]  &&
	local x tgt &&
	{
		x="${FUNCNAME[1]#__fullup_configure_}"
		tgt="${x#*_}"
		[[ "$tgt" && ( "$tgt" != "$x" ) ]]
	} && {
		x="${x%_*}"
		[[ "$x" ]] &&
		[[ "__fullup_configure_${x}_$tgt" == "${FUNCNAME[1]}" ]]
	} && {
		x="__fullup_configure_${1}_$tgt"

		# If there's no function with our constructed name we consider
		# this a successful end to the dependency chain walk. That way
		# we can write e.g. "__configure end" and not worry about it.
		declare -F "$x" >/dev/null || {
			return 0
		}

		"$x"
		return
	} || exit 1
}

### include guard
} || ((1))
###
