### include guard
(return 0 2>/dev/null)          &&
[[ -z ${FULLUP_COMMON_SH_+1} ]] &&
declare -xr FULLUP_COMMON_SH_=1 && {
###

. "$I/log.sh"
. "$I/dev.sh"

die() { err "$@"; exit 1; }

not_implemented() {
	err '%s not implemented yet' "${FUNCNAME[1]}"
}

prompt() {
	local d="${1,}"
	local s="${2:+$2 }"

	case "${d,}" in
	n) s+='[y/N] '; d='1' ;;
	y) s+='[Y/n] '; d='0' ;;
	*) s+='[y/n] '; d=''  ;;
	esac

	printf %s "$s" >&2

	if (( __fullup_cfg_no_prompt )); then
		return $d
	fi

	s=${d:+return ${d}}

	while read -r -s -N1; do
		case "$REPLY" in
		N|n) return 1 ;;
		Y|y) return 0 ;;
		$'\n') ${s}   ;;
		esac
	done

	die "'read -r -s -N1' failed"
}

check_md5() {
	local err_msg
	local lineno
	local -i retval

	lineno="$(cut -b33- "$1" | grep -Fxn "  $2")"
	retval="$?"
	lineno="${lineno%%:*}"

	if (( retval != 0 || lineno <= 0 )); then
		err 'missing MD5 sum: %q' "$2"
	elif ! { cut -d$'\n' -f"$lineno" "$1" | md5sum --status -c -; }; then
		err 'MD5 check error: %q' "$2"
	else
		return 0
	fi

	(( __fullup_cfg_ignore_md5 ))
}

check_cpu() {
	not_implemented
}

if [[ "$HAVE_SPL" ]]; then
install_spl() {
	not_implemented
}
fi

if [[ "$HAVE_BOOTLOADER" ]]; then
install_bootloader() {
	local bdev="${__fullup_cfg_bootloader_dev}"
	local -a dd_args=(
		"if=${__fullup_cfg_bootloader_src}"
		"of=/dev/$bdev"
		${__fullup_cfg_bootloader_dd_bs:+ \
			"bs=${__fullup_cfg_bootloader_dd_bs}" }
		${__fullup_cfg_bootloader_dd_seek:+ \
			"seek=${__fullup_cfg_bootloader_dd_seek}" }
	)

	msg "Installing bootloader"

	${dry_eval_}echo '0' '>' "/sys/block/$bdev/force_ro"
	${dry_}dd "${dd_args[@]}" && ${dry_}sync
	#${dry_eval_}echo '1' '>' "/sys/block/$bdev/force_ro"
}
fi

format_ext4_fs() {
	dev_is_valid "$1" || {
		err 'invalid device %q' "$1"
		return 1
	}

	dev_umount "$1" || {
		wrn 'unmounting %q failed' "$1"
	}

	msg 'Formatting ext4 filesystem on %q' "$1"

	dev_mkfs_ext4 "$@"
}

wipe_user_data() {
	msg 'Wiping user partition'
	format_ext4_fs "$__fullup_cfg_user_part"
}

install_rootfs() {
	msg 'Installing root filesystem'

	format_ext4_fs "$__fullup_cfg_root_part" 'root' &&
	dev_mount      "$__fullup_cfg_root_part"        &&
	{
		local r
		local busybox_tar="$(tar 2>&1|grep BusyBox)"

		r="$(realpath -e "$__fullup_cfg_rootfs_src")" &&
		if command -v pv &> /dev/null; then
			# Check if parallel compressor available, multiple times faster decompression
			# Busybox tar doesn't support custom compressors.
			if [[ ! "$busybox_tar" ]] && command -v pbzip2 &> /dev/null; then
				${dry_eval_}pv --force "$r" '2>' /dev/tty1 '|' tar -I pbzip2 -x -C "/mnt/$__fullup_cfg_root_part"
			else
				${dry_eval_}pv --force "$r" '2>' /dev/tty1 '|' tar -xj -C "/mnt/$__fullup_cfg_root_part"
			fi
		else
			${dry_}tar -xf "$r" -C "/mnt/$__fullup_cfg_root_part"
		fi

		r="$?"

		dev_umount "$__fullup_cfg_root_part" || {
			wrn 'unmounting %q failed' "$__fullup_cfg_root_part"
		}

		return "$r"
	}
}

cp_to_mnt() {
	local -r mnt="/mnt/$1"
	local -r -n src="__fullup_cfg_$2"
	local -r -n dest="__fullup_cfg_$3"
	local -i c=${#src[@]} n=${#dest[@]} i
	local -r f="${FUNCNAME[0]}: "

	(( n == c )) || {
		wrn "${f}array sizes differ (%s=%d %s=%d)" "$2" "$c" $3 "$n"
		(((n < c) || (n = c)))
	}

	(( n > 0 )) || {
		err "${f}nothing to copy"
		return 1
	}

	for ((i = 0, c = 0; i < n; i++)); do
		if [[ -z "${src[i]}" ]]; then
			err "${f}null %s%${dest[i]:+.}s" "$2[$i]" " and $3[$i]"
		elif [[ -z "${dest[i]}" ]]; then
			err "${f}null %s" "$3[$i]"
		elif ! ${dry_}cp "${src[i]}" "$mnt/${dest[i]}"; then
			err "${f}error copying %q" "${src[i]}"
		else
			msg 'Installed %q on %q as %q' \
			    "${src[i]}" "$1" "${dest[i]}"
			((c++))
		fi
	done

	(( c > 0 )) || {
		err "${f}nothing copied"
		return 1
	}

	(( c == n ))
}

mount_and_copy_files() {
	dev_mount "$1" || {
		err "${FUNCNAME[0]}: can't mount %q" "$1"
		return 1
	}
	cp_to_mnt "$@"
}

maybe_install_boot_part() {
	local dev="$__fullup_cfg_boot_part"
	local -i tried=0 gave_up=0

	if (( ! __fullup_cfg_no_kernel )); then
		((tried++))
		newline
		msg "Installing kernel image"
		mount_and_copy_files "$dev" kernel_src kernel_dest || {
			newline
			confirm 'failed to install kernel, continue anyway?'
			gave_up="$?"
		}
	fi

	if [[ "$HAVE_DTB" ]] &&
	   (( ! gave_up && ! __fullup_cfg_no_dtb )); then
		((tried++))
		newline
		msg "Installing device tree blobs"
		mount_and_copy_files "$dev" dtb_src dtb_dest || {
			newline
			confirm 'failed to install device trees, continue anyway?'
			gave_up="$?"
		}
	fi

	if (( tried )); then
		dev_umount "$dev"
	fi

	return "$gave_up"
}

maybe_reboot() {
	case "${__fullup_cfg_reboot}" in
	'main') ;;
	'no')
		msg 'Not rebooting'
		return 0
		;;
	'rescue')
		${dry_eval_}echo "'1 0'" '>' '/dev/factory_reset'
		;;
	*)
		die 'invalid reboot target %q' "${__fullup_cfg_reboot}"
		;;
	esac

	# Give some time for the user to observe the log before rebooting
	local t=5
	msg "Rebooting to %s${t:+ in %d seconds}" "${__fullup_cfg_reboot}" "$t"
	sleep "$((t))"

	${dry_}systemctl reboot
}

### include guard
} || ((1))
###
